@@ -1,10 +1,13 @@
 package Apache::Status;
 use strict;
+eval { use Devel::Symdump (); };
 
 my(%status) = (
    inc => "Loaded Modules",
    rgysubs => "Compiled Registry Scripts",
    symdump => "Symbol Table Dump",
+   inh_tree => "Inheritance Tree",
+   isa_tree => "ISA Tree",	       
 );
 
 sub menu_item {
@@ -55,7 +58,6 @@ EOF
 
 sub symdump {
     my($package) = @_;
-    require Devel::Symdump;
     my $sob = Devel::Symdump->rnew($package);
     return $sob->as_HTML($package);
 }
@@ -99,6 +101,9 @@ sub status_rgysubs {
     \@retval;
 }
 
+sub status_inh_tree { ["<pre>", Devel::Symdump->inh_tree, "</pre>"] }
+sub status_isa_tree { ["<pre>", Devel::Symdump->isa_tree, "</pre>"] }
+
 1;
 
 __END__
